/*
 * NAME:
 *	ssl - routines useful for SSL servers and clients
 *
 * DESCRIPTION:
 *	These routines provide for client authentication using
 *	Eric Young's SSLeay library.
 *
 * SEE ALSO:
 *	sslfd.h, sslc.c, ssld.c, rwnet.c
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: ssl.c,v 1.28 1997/05/17 12:55:14 sjg Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <string.h>


#include "x509.h"
#include "ssl.h"

#define _EXTERN
#include "sslfd.h"

/* make sure we get cipher's we trust */
#ifndef SSL_CIPHER_LIST
# define SSL_CIPHER_LIST "RC4-MD5:IDEA-CBC-MD5:DES-CBC3-MD5:DES-CBC-MD5"
#endif
#ifndef SSL_VERIFY_FLAG
# define SSL_VERIFY_FLAG SSL_VERIFY_PEER
#endif
#ifndef NUMFDS
# define NUMFDS 64
#endif

static SSL_CTX	*ssl_ctx;
static SSL	*ssl_con;
char x509_name[256];

/* these are set by ssl_setopt() */
static char *ssl_cert_file = 0;
static char *ssl_key_file = 0;
static char *ssl_cipher_list = 0;

static SSL *ssl_cons[NUMFDS];

int
ssl_setcon(fd, con)
	int fd;
	SSL *con;
{
	static int once = 1;	

	if (once) {
		memset((char *)ssl_cons, 0, sizeof(ssl_cons));
		once = 0;
	}
	if (fd >= 0 && fd < NUMFDS) {
		ssl_cons[fd] = con;
		return 1;
	}
	return 0;
}

SSL *
ssl_getcon(fd)
	int fd;
{
	ssl_setcon(-1, 0);			/* ensure initialized */
	if (fd >= 0 && fd < NUMFDS) {
		return ssl_cons[fd];
	}
	return 0;
}

/* old names */
int
set_sslcon(fd, con)
	int fd;
	SSL *con;
{
	return ssl_setcon(fd, con);
}

SSL *
get_sslcon(fd)
	int fd;
{
	return ssl_getcon(fd);
}

int
ssl_pending(fd)
	int fd;
{
	SSL *con;
	
	if (con = ssl_getcon(fd)) {
		return SSL_pending(con);
	}
	return 0;
}


int
ssl_close(fd)
	int fd;
{
	SSL *con;
	int i;

	if (con = ssl_cons[fd]) {
		if (con != ssl_con) {		/* want to keep this one */
			if (fd < 3) {
				/*
				 * stdin,stdout,stderr often refer to
				 * same socket, avoid free()ing twice.
				 */
				for (i = 0; i < 3; i++) {
					if (ssl_cons[i] == con)
						ssl_cons[i] = 0;
				}
			}
			SSL_free(con);
		}
		ssl_setcon(fd, 0);
	}
	return close(fd);
}

void
ssl_err(log_lev, fmt)
	int log_lev;
	char *fmt;
{
	if (log_lev > 0) {
		syslog(log_lev, fmt, ERR_error_string(ERR_get_error(), NULL));
	} else if (log_lev < 0) {
		fprintf(stderr, fmt, ERR_error_string(ERR_get_error(), NULL));
	}
}

/*
 * this is mainly here for telnet's use
 */
int
ssl_get_pw(key_file, log_lev)
	const char *key_file;
	int log_lev;
{
	if (!key_file)
		key_file = ssl_key_file;
	
	if (key_file && *key_file) {
		if (SSL_use_RSAPrivateKey_file(ssl_con,
					       (char *) key_file,
					       X509_FILETYPE_PEM)<=0) {
			ssl_err(log_lev, "SSL_use_RSAPrivateKey_file error %s\n");
			return (-1);
		}
		if (key_file == ssl_key_file) {
			free(ssl_key_file);
			ssl_key_file = 0;
		}
	}
	return 0;
}

int
ssl_init(cert_file, key_file, cipher_list, verify_flag, verify_cb, log_lev, flags)
	const char *cert_file, *key_file, *cipher_list;
	int verify_flag;
	int (*verify_cb)();
	int log_lev;
	int flags;
{
	int sts = -1;
	char *err = 0;

	if (!cert_file)
		cert_file = ssl_cert_file;
	if (!cert_file)
		cert_file = getenv("SSL_CERT");

	if (!key_file)
		key_file = ssl_key_file;
	if (!key_file)
		key_file = getenv("SSL_KEY");
	if (ssl_cipher_list)
		cipher_list = ssl_cipher_list;
	if (!cipher_list)
		cipher_list = getenv("SSL_CIPHER");
#ifdef SSL_CIPHER_LIST
	if (!cipher_list)
		cipher_list = SSL_CIPHER_LIST;
#endif
	if (verify_flag >= 0)
		ssl_verify_flag = verify_flag;
	else if (ssl_verify_flag < -1)
		ssl_verify_flag = SSL_VERIFY_FLAG|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
	else if (ssl_verify_flag < 0)
		ssl_verify_flag = SSL_VERIFY_FLAG;

	if (ssl_key_file) {
		free(ssl_key_file);
		ssl_key_file = 0;
	}
	if (log_lev > 0)
		syslog(LOG_DEBUG, "cert_file=%s,key_file=%s,cipher_list=%s",
		       cert_file ? cert_file : "",
		       key_file ? key_file : "",
		       cipher_list ? cipher_list : ""
		       );
	
	SSL_load_error_strings();
	if ((ssl_ctx = SSL_CTX_new()) == 0) {
		ssl_err(log_lev, "cannot create SSL context: %s\n");
		return (-1);
	}
		
	if (!X509_set_default_verify_paths(ssl_ctx->cert)) {
		ssl_err(log_lev, "cannot load certificates via X509_set_default_verify_paths %s\n");
		return (-1);
	}
	ssl_con = SSL_new(ssl_ctx);
	if (cert_file && *cert_file) {
		if (!(key_file && *key_file))
			key_file = cert_file;

		if (SSL_use_certificate_file(ssl_con, (char *) cert_file,
					     X509_FILETYPE_PEM)<=0) {
			ssl_err(log_lev, "SSL_use_certificate_file error %s\n");
			return (-1);
		}
		if (key_file) {
			if (flags & 2)		/* defer key read */
				ssl_key_file = strdup(key_file);
			else {
				if (ssl_get_pw(key_file, log_lev) < 0) {
					return (-1);
				}
			}
		}
	} else {
		if (log_lev > 0)
			syslog(log_lev, "no cert_file");
	}
		
	SSL_set_verify(ssl_con, ssl_verify_flag, verify_cb);
	if (cipher_list && *cipher_list)
		SSL_set_cipher_list(ssl_con,(char *) cipher_list);
	return (cert_file != 0);
}

/*
 * this does the real work for ssl_{start,clone}*()
 */
int
ssl_start1(s, log_lev, active, con, cb, vf)
	int s;
	int log_lev;
	int active;
	SSL	*con;
	int (*cb) ();
	int vf;
{
	int sts = -1;
	
	if (con == 0) {
		if (ssl_con == 0) {
			if (ssl_init(0, 0, 0, -1, cb, log_lev, 1) < 0)
				return -1;
		}
		if (ssl_ctx && (con = SSL_new(ssl_ctx))) {
			/*
			 * This just picks up details of our own
			 * cert etc...
			 */
			memcpy((char *)con, (char *)ssl_con, sizeof(SSL));
		}
	}	
	if (con == 0) {
		return -1;
	}
	SSL_set_fd(con,s);

	/*
	 * <sjg> we defer reading key until actually needed
	 */
	if (ssl_key_file) {
		/*
		 * NOTE: ssl_get_pw() assumes con==ssl_con
		 */
		if (ssl_get_pw(ssl_key_file, log_lev) < 0) {
			return (-1);
		}
	}
	/*
	 * <sjg> this is confusing, but the acceptor always provides a
	 * cert, where as it is optional for the connector.
	 * In cases  were the application only cares about client
	 * authentication, we can reverse the SSL roles to cut down
	 * the startup overheads.
	 * Mutual auth with 512bit keys takes up to 4s on a i486DX33
	 */
	if (cb) {
		SSL_set_verify(con, (vf < 0) ? ssl_verify_flag : vf, cb);
	}
	if (active) {
		sts = SSL_connect(con);
	} else {
		sts = SSL_accept(con);
	}
	if (sts <= 0) {
		if (log_lev > 0) {
			syslog(log_lev, "SSL_%s error %s",
			       (active) ? "connect" : "accept",
			       ERR_error_string(ERR_get_error(), NULL));
		} else if (log_lev < 0) {
			fprintf(stderr, "SSL_%s error %s\n",
				(active) ? "connect" : "accept",
				ERR_error_string(ERR_get_error(), NULL));
		}
		return (-1);
	}
	ssl_setcon(s, con);
	return sts;
}

int
ssl_start_cb(s, log_lev, active, cb, vf)
	int s;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
{
	int sts;
	
	if ((sts = ssl_start1(s, log_lev, active, 0, cb, vf)) > 0)
		ssl_fd = s;
	return sts;
}


int
ssl_start(s, log_lev, active)
	int s;
	int log_lev;
	int active;
{
	return ssl_start_cb(s, log_lev, active, 0, -1);
}


int
ssl_clone_cb(s, old, log_lev, active, cb, vf)
	int s;
	int old;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
{
	SSL *con, *con2 = SSL_new(ssl_ctx);
	int sts = -1;
	
	if (con2) {
		if (old < 0)
			old = ssl_fd;	/* last primary session started */
		if (con = ssl_getcon(old)) {
			SSL_copy_session_id(con2, con);
			sts = ssl_start1(s, log_lev, active, con2, cb, vf);
		} else
			SSL_free(con2);
	}
	return sts;
}


int
ssl_clone(s, old, log_lev, active)
	int s;
	int old;
	int log_lev;
	int active;
{
	return ssl_clone_cb(s, old, log_lev, active, 0, -1);
}

/*
 * some old names
 */
int
ssl_start2_cb(s, old, log_lev, active, cb, vf)
	int s;
	int old;
	int log_lev;
	int active;
	int (*cb) ();
	int vf;
{
	return ssl_clone_cb(s, old, log_lev, active, cb, vf);
}

int
ssl_start2(s, log_lev, active)
	int s;
	int log_lev;
	int active;
{
	return ssl_clone_cb(s, -1, log_lev, active, 0, -1);
}

char *
get_x509_name(buf, bufsz)
	char *buf;
	int bufsz;
{
	char *p, *e;
	int n = 0;

	if (buf == 0) {
		buf = x509_name;
		bufsz = sizeof(x509_name);
	}
	if (ssl_auth_cert) {
		if (p = strstr(ssl_auth_cert, "/CN=")) {
			p += 4;
			if (e = strpbrk(p, "/\r\n")) {
				n = e - p;
				if (n >= bufsz)
					n = bufsz - 1;
			} else
				n = bufsz - 1;
			(void) strncpy(buf, p, n);
			buf[n] = '\0';
		}
		return buf;
	}
	return 0;
}

#define OPTARG_VAL(str, cp, key) ((cp = strchr(str, '=')) && strncmp(str, key, cp - str) == 0 && ++cp)

int
ssl_setopt(optarg)
	const char *optarg;
{
	static int use_ssl = 1;
	char *cp;
	
	if (OPTARG_VAL(optarg, cp, "cert"))
		ssl_cert_file = cp;
	else if (OPTARG_VAL(optarg, cp, "cipher"))
		ssl_cipher_list = cp;
	else if (OPTARG_VAL(optarg, cp, "key"))
		ssl_key_file = cp;
	else if (OPTARG_VAL(optarg, cp, "verify"))
		ssl_verify_flag = atoi(cp);
	else if (OPTARG_VAL(optarg, cp, "auth"))
		_path_ssl_users = cp;
	else if (strcmp(optarg, "ssl") == 0 || strcmp(optarg, "ssl_only") == 0)
		use_ssl |= 4;
	else if (strcmp(optarg, "secure") == 0) {
		use_ssl |= 2;
		ssl_secure_flag = 1;
	} else if (strcmp(optarg, "certsok") == 0 )
		ssl_certsok_flag = 1;
	return use_ssl;
}
